import json
import os
import re
import time
import traceback
import uuid
from contextlib import nullcontext
from urllib.request import pathname2url

from .. import util
from .._polyfill import zipfile
from ..util import Info
from .host import Host

REGEX_TARGET_FILENAME_FORMATTER = re.compile(r'%([^%]*)%')


class Importer():
    """Main class for importing."""
    def __init__(self, book, *,
                 target_id=None,
                 target_index=None,
                 target_filename=None,
                 rebuild_folders=False,
                 prune=False,
                 resolve_id_used='skip',  # skip, replace, new
                 ):
        self.book = book
        self.target_id = target_id
        self.target_index = target_index if (isinstance(target_index, int) and target_index >= 0) else None
        self.target_filename = target_filename or '%ID%'
        self.rebuild_folders = rebuild_folders
        self.prune = prune
        self.resolve_id_used = resolve_id_used

    def run(self, files=None):
        self.book.load_meta_files()
        self.book.load_toc_files()

        self.map_eid_to_info = {}
        self.map_id_to_new_id = {}

        book_meta_orig = self.book.checksum(self.book.meta)
        book_toc_orig = self.book.checksum(self.book.toc)

        # fix target_id
        if not self.rebuild_folders:
            if self.target_id is None:
                self.target_id = self.book.ROOT_ITEM_ID
            elif not (self.target_id in self.book.meta or self.target_id in self.book.SPECIAL_ITEM_ID):
                yield Info('warn', f'Target ID {self.target_id!r} is invalid. Use {self.book.ROOT_ITEM_ID!r} instead.')
                self.target_id = self.book.ROOT_ITEM_ID

        for file in files:
            yield Info('info', f'Importing archive file {os.path.basename(file)!r}')
            try:
                successful = yield from self._import_file(file)
            except Exception as exc:
                # unexpected error
                traceback.print_exc()
                yield Info('error', f'Failed to import file {os.path.basename(file)!r}: {exc}', exc=exc)
            else:
                # finalize a successfully imported file
                if successful and self.prune:
                    yield Info('debug', f'Removing {os.path.basename(file)!r} (prune)')
                    os.remove(file)

        # update files
        if self.book.checksum(self.book.meta) != book_meta_orig:
            yield Info('info', 'Saving changed meta files...')
            self.book.save_meta_files()

        if self.book.checksum(self.book.toc) != book_toc_orig:
            yield Info('info', 'Saving changed TOC files...')
            self.book.save_toc_files()

    def generate_imported_filename(self, id, meta, export_info):
        """Generate an adequate filename (without file extension) for an
        importing item.
        """
        def date_formatter(id, pattern):
            if pattern == '':
                return id

            dt = util.id_to_datetime(id)

            if not dt:
                return ''

            if pattern == 'UTC_DATE':
                return f'{dt.year:04d}-{dt.month:02d}-{dt.day:02d}'

            if pattern == 'UTC_TIME':
                return f'{dt.hour:02d}-{dt.minute:02d}-{dt.second:02d}'

            if pattern == 'UTC_YEAR':
                return f'{dt.year:04d}'

            if pattern == 'UTC_MONTH':
                return f'{dt.month:02d}'

            if pattern == 'UTC_DAY':
                return f'{dt.day:02d}'

            if pattern == 'UTC_HOURS':
                return f'{dt.hour:02d}'

            if pattern == 'UTC_MINUTES':
                return f'{dt.minute:02d}'

            if pattern == 'UTC_SECONDS':
                return f'{dt.second:02d}'

            ldt = dt.astimezone()

            if pattern == 'DATE':
                return f'{ldt.year:04d}-{ldt.month:02d}-{ldt.day:02d}'

            if pattern == 'TIME':
                return f'{ldt.hour:04d}-{ldt.minute:02d}-{ldt.second:02d}'

            if pattern == 'YEAR':
                return f'{ldt.year:04d}'

            if pattern == 'MONTH':
                return f'{ldt.month:02d}'

            if pattern == 'DAY':
                return f'{ldt.day:02d}'

            if pattern == 'HOURS':
                return f'{ldt.hour:02d}'

            if pattern == 'MINUTES':
                return f'{ldt.minute:02d}'

            if pattern == 'SECONDS':
                return f'{ldt.second:02d}'

            return ''

        def formatter(m):
            key = m.group(1)

            if key == '':
                return '%'

            if key == 'ID':
                return id

            if key == 'EID':
                return export_info['id']

            if key == 'UUID':
                return str(uuid.uuid4())

            if key == 'TITLE':
                return meta.get('title', '')

            if key == 'SOURCE':
                return meta.get('source', '')

            key, _, pattern = key.partition(':')

            if key == 'CREATE':
                return str(date_formatter(meta.get('create', ''), pattern))

            if key == 'MODIFY':
                return str(date_formatter(meta.get('modify', ''), pattern))

            if key == 'EXPORT':
                return str(date_formatter(export_info['timestamp'], pattern))

            return ''

        filename = REGEX_TARGET_FILENAME_FORMATTER.sub(formatter, self.target_filename)
        filename = '/'.join(util.validate_filename(s) for s in filename.split('/'))
        return filename

    def _import_file(self, file):
        successful = True
        with zipfile.ZipFile(file) as zh:
            topdirs = set()
            for zinfo in zh.infolist():
                topdir, _, _ = zinfo.filename.partition('/')
                topdirs.add(topdir)

            for topdir in sorted(topdirs):
                yield Info('debug', f'Importing entry {topdir!r}')
                try:
                    yield from self._import_topdir(zh, topdir)
                except RuntimeError as exc:
                    # intended raise to skip the import
                    yield Info('error', f'Failed to import entry {topdir!r}: {exc}', exc=exc)
                    successful = False
                except Exception as exc:
                    # unexpected error
                    traceback.print_exc()
                    yield Info('error', f'Failed to import entry {topdir!r}: {exc}', exc=exc)
                    successful = False

        return successful

    def _import_topdir(self, zh, topdir):
        try:
            with zh.open(f'{topdir}/export.json') as fh:
                export_info = json.load(fh)
        except Exception as exc:
            raise RuntimeError(f"Unable to read 'export.json': {exc}") from exc

        if export_info['version'] == 2:
            try:
                assert isinstance(export_info['id'], str)
                assert isinstance(export_info['timestamp'], str)
                assert isinstance(export_info['timezone'], int)
                assert isinstance(export_info['path'], list)
                assert isinstance(export_info['index'], int)
            except (AssertionError, KeyError) as exc:
                raise RuntimeError("Malformed 'export.json'") from exc

        else:
            raise RuntimeError(f'Unsupported archive version: {export_info["version"]!r}')

        try:
            with zh.open(f'{topdir}/meta.json') as fh:
                meta = json.load(fh)
        except Exception as exc:
            raise RuntimeError(f"Unable to read 'meta.json': {exc}") from exc

        id = meta.pop('id')
        if id in self.book.SPECIAL_ITEM_ID:
            raise RuntimeError(f'invalid ID {id!r}')

        # skip importing data for a duplicated occurrence of a previously
        # imported item
        imported_id = self.map_eid_to_info.setdefault(export_info['id'], {}).get('id')
        if imported_id is not None:
            id = imported_id
            yield Info('debug', f'Skipped importing data for multi-referenced {id!r}')
        else:
            id = yield from self._import_meta_and_data(id, meta, zh, topdir, export_info)

        parent_id = yield from self._insert_to_toc(id, export_info)
        text_parent = '' if parent_id is None else f' (under {parent_id!r})'
        yield Info('info', f'Imported {id!r}{text_parent}')
        self.map_eid_to_info.setdefault(export_info['id'], {}).setdefault('id', id)

    def _import_meta_and_data(self, id, meta, zh, topdir, export_info):
        """Import meta and data

        Returns:
            string: ID of the imported item
        """
        index = meta.get('index', '')

        if index:
            if index.endswith('/index.html'):
                src = f'{topdir}/data/{os.path.dirname(index)}'
            else:
                src = f'{topdir}/data/{index}'

            # determine normal copy dst
            _, ext = os.path.splitext(src)
            filename = self.generate_imported_filename(id, meta, export_info) + ext
            dst = os.path.normpath(os.path.join(self.book.data_dir, filename))
            meta['index'] = filename + ('/index.html' if index.endswith('/index.html') else '')

        # handle resolve cases if id exists
        # may overwrite id, dst, and meta['index']
        new_id = id
        if id in self.book.meta:
            if self.resolve_id_used == 'skip':
                raise RuntimeError(f'ID {id!r} already exists')

            elif self.resolve_id_used == 'replace':
                index_old = self.book.meta[id].get('index', '')

                # replace only if index type matches
                if os.path.splitext(index)[1] != os.path.splitext(index_old)[1]:
                    raise RuntimeError('index type not match')

                if index_old.endswith('/index.html') != index.endswith('/index.html'):
                    raise RuntimeError('index type not match')

                yield Info('warn', f'Importing duplicated {id!r} in place of the current...')

                if index:
                    # use original index
                    meta['index'] = index_old

                    # remove current index file or folder
                    if index.endswith('/index.html'):
                        dst = os.path.normpath(os.path.join(self.book.data_dir, os.path.dirname(index_old)))
                    else:
                        dst = os.path.normpath(os.path.join(self.book.data_dir, index_old))
                    try:
                        util.fs.delete(dst)
                    except util.fs.FSEntryNotFoundError:
                        pass

                self.map_eid_to_info.setdefault(export_info['id'], {}).setdefault('replaced', True)

            elif self.resolve_id_used == 'new':
                new_id = self.book.get_unique_id()
                yield Info('warn', f'Importing duplicated {id!r} as {new_id!r}...')

                if index:
                    # overwrite dst and index
                    filename = self.generate_imported_filename(new_id, meta, export_info) + ext
                    dst = os.path.normpath(os.path.join(self.book.data_dir, filename))
                    meta['index'] = filename + ('/index.html' if index.endswith('/index.html') else '')

            else:
                raise RuntimeError(f'unknown resolve mode: {self.resolve_id_used!r}')

        # if a new folder for id has been generated, replace it with new_id
        # (e.g. when X/Y has been imported before Z/X, there will be an X' when
        # importing X under Z)
        try:
            folder_id = self.map_id_to_new_id[id]
        except KeyError:
            pass
        else:
            yield Info('info', f'Replacing {folder_id!r} with {new_id!r}')
            try:
                self.book.toc[new_id] = self.book.toc.pop(folder_id)
            except KeyError:
                # folder_id not in toc
                pass

            # @TODO: better algorithm for the global replacement
            for toc in self.book.toc.values():
                for i, v in enumerate(toc):
                    if v == folder_id:
                        toc[i] = new_id

            del self.book.meta[folder_id]

        # mark id as generated
        self.map_id_to_new_id[id] = new_id

        # import data files
        if index:
            if os.path.lexists(dst):
                raise RuntimeError(f'file {dst!r} already exists')

            yield Info('debug', f'Extracting data files to {self.book.get_subpath(dst)!r}')
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            util.fs.zip_extract(zh, dst, src, tzoffset=export_info['timezone'])

        # import favicon
        for f in zh.namelist():
            if f.startswith(f'{topdir}/favicon/') and not f.endswith('/'):
                basename = os.path.basename(f)
                iconfile = os.path.join(self.book.tree_dir, 'favicon', basename)
                os.makedirs(os.path.dirname(iconfile), exist_ok=True)

                try:
                    util.fs.zip_extract(zh, iconfile, f, tzoffset=export_info['timezone'])
                except FileExistsError:
                    yield Info('debug', f'Skipped existing favicon cache {basename!r}')
                else:
                    yield Info('info', f'Added favicon cache {basename!r}')

                # rewrite icon property to be consistent with the importing book
                try:
                    base = dst if index.endswith('/index.html') else os.path.dirname(dst)
                except UnboundLocalError:
                    base = self.book.data_dir
                meta['icon'] = pathname2url(os.path.relpath(iconfile, base))

                break

        self.book.meta[new_id] = meta
        return new_id

    def _insert_to_toc(self, id, export_info):
        """Insert the importing item to TOC

        Returns:
            string: ID of the parent the item is inserted under
        """
        if self.map_eid_to_info.setdefault(export_info['id'], {}).get('replaced'):
            yield Info('debug', f'Skipped inserting replaced {id!r}')
            return None

        # deduplicate by checking the ref_key
        if self.rebuild_folders:
            export_path = export_info['path']
            ref_key = (export_path[-1]['id'], export_info['index'])
        else:
            ref_key = (self.target_id, None)

        if ref_key in self.map_eid_to_info[export_info['id']].setdefault('refs', set()):
            yield Info('debug', f'Skipped inserting multi-referenced {id!r}')
            return None

        self.map_eid_to_info[export_info['id']]['refs'].add(ref_key)

        # perform the insertion
        if not self.rebuild_folders:
            parent_id = self.target_id
            self._insert_to_id(id, parent_id)
            return parent_id

        if self.target_id is None:
            for i in reversed(range(len(export_path))):
                parent_id = export_path[i]['id']
                if parent_id in self.book.SPECIAL_ITEM_ID:
                    break

                parent_id = self.map_id_to_new_id.get(parent_id, parent_id)

                # Assume that an item in meta also exists somewhere in toc, except
                # for id, which is the one being imported and is not in toc (yet).
                if parent_id == id:
                    continue

                if parent_id in self.book.meta:
                    break
            else:
                # for a bad path data not starting from 'root'
                i = -1
                parent_id = self.book.ROOT_ITEM_ID
        else:
            for i in reversed(range(len(export_path))):
                parent_id = export_path[i]['id']
                if parent_id in self.book.SPECIAL_ITEM_ID:
                    parent_id = self.target_id
                    break

                parent_id = self.map_id_to_new_id.get(parent_id)

                if parent_id == id:
                    continue

                if parent_id in self.book.meta:
                    break
            else:
                # for a bad path data not starting from 'root'
                i = -1
                parent_id = self.target_id

        for j in range(i + 1, len(export_path)):
            folder_id = export_path[j]['id']
            folder_title = export_path[j]['title']

            # special handling for id
            if folder_id == id:
                new_id = folder_id
                if parent_id == self.map_id_to_new_id.get(export_path[-1]['id'], export_path[-1]['id']):
                    # this ancestor is identical to the direct parent,
                    # and folder_id will eventually be inserted under it
                    yield Info('debug', f'Skipped inserting {new_id!r} under {parent_id!r} (same as parent)')
                elif new_id not in self.book.toc.get(parent_id, ()):
                    self._insert_to_id(new_id, parent_id, allow_insert=parent_id == self.target_id)
                    yield Info('info', f'Inserted {new_id!r} under {parent_id!r}')
                else:
                    yield Info('debug', f'Skipped inserting {new_id!r} under {parent_id!r} (already in)')
            else:
                try:
                    new_id = self.map_id_to_new_id[folder_id]
                except KeyError:
                    new_items = self.book.add_item({
                        'title': folder_title,
                        'type': 'folder',
                    }, None)
                    new_id = next(iter(new_items))
                    self.map_id_to_new_id[folder_id] = new_id
                    self._insert_to_id(new_id, parent_id, allow_insert=parent_id == self.target_id)
                    yield Info('info', f'Generated folder {new_id!r} for missing {folder_id!r} under {parent_id!r}')
                else:
                    if new_id not in self.book.toc.get(parent_id, ()):
                        self._insert_to_id(new_id, parent_id, allow_insert=parent_id == self.target_id)
                        yield Info('info', f'Inserted folder {new_id!r} under {parent_id!r}')
                    else:
                        yield Info('debug', f'Skipped inserting {new_id!r} under {parent_id!r} (already in)')

            parent_id = new_id

        self._insert_to_id(id, parent_id, allow_insert=parent_id == self.target_id)
        return parent_id

    def _insert_to_id(self, id, parent_id, allow_insert=True):
        parent = self.book.toc.setdefault(parent_id, [])

        if allow_insert and self.target_index is not None:
            parent.insert(self.target_index, id)
            if not self.book.config['new_at_top']:
                self.target_index += 1
        else:
            target_index = 0 if self.book.config['new_at_top'] else len(parent)
            parent.insert(target_index, id)


def run(host, files, book_id='', *, lock=True, **kwargs):
    start = time.time()

    if isinstance(host, Host):
        pass
    elif isinstance(host, str):
        host = Host(host)
    else:
        host = Host(*host)

    # Fail for invalid book ID
    if book_id not in host.books:
        yield Info('error', f'Invalid book {book_id!r}.')
        return

    yield Info('debug', f'Loading book {book_id!r}.')

    try:
        book = host.books[book_id]

        if book.no_tree:
            yield Info('error', f'Unable to import to book {book_id!r} ({book.name!r}) (no_tree).')
            return

        yield Info('info', f'Impoting to book {book_id!r} ({book.name!r}).')
        lh = book.get_tree_lock(persist=lock).acquire() if lock else nullcontext()
        with lh:
            generator = Importer(book, **kwargs)
            yield from generator.run(files)

    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
        return
    else:
        yield Info('info', 'Done.')

    yield Info('info', '----------------------------------------------------------------------')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



"""Generator of integrity check for scrapbook data.
"""
import os
import time
import traceback
from contextlib import nullcontext
from datetime import datetime
from urllib.parse import unquote, urlsplit

from .. import WSB_DIR, util
from .._polyfill import zipfile
from ..util import Info
from .book import TreeFileError
from .host import Host
from .indexer import (
    SUPPORT_FOLDER_SUFFIXES,
    FavIconCacher,
    Indexer,
    generate_item_create,
    generate_item_modify,
)

# threshold (in seconds) to report file mtime newer than item modify property
# more than a normal page capture delay
OLDER_MTIME_THRESHOLD = 10 * 60


class BookChecker:
    RESOLVE_METHODS = {
        'resolve_invalid_id',
        'resolve_missing_index',
        'resolve_missing_index_file',
        'resolve_missing_date',
        'resolve_older_mtime',
        'resolve_toc_unreachable',
        'resolve_toc_invalid',
        'resolve_toc_empty_subtree',
        'resolve_unindexed_files',
        'resolve_absolute_icon',
        'resolve_unused_icon',
    }

    def __init__(self, book, resolve_all=False, **kwargs):
        self.book = book

        # init resolve* arguments
        self.resolve = False
        for k in self.RESOLVE_METHODS:
            if not resolve_all:
                v = kwargs.get(k)
                if not v:
                    setattr(self, k, False)
                    continue
            setattr(self, k, True)
            self.resolve = True

        self.wsb_dir = os.path.join(book.root, WSB_DIR)
        self.book_wsb_dir = os.path.join(book.top_dir, WSB_DIR)

    def run(self):
        self.seen_in_toc = set()
        self.find_index_exclude = set()
        self.used_favicons = set()

        self.cnt_warns = 0
        self.cnt_errors = 0
        self.cnt_resolves = 0
        self.cnt_items = 0
        self.cnt_dirs = 0
        self.cnt_files = 0
        self.cnt_bytes = 0

        yield Info('info', 'Loading tree...')
        self._load_tree()

        book_meta_orig = self.book.checksum(self.book.meta)
        book_toc_orig = self.book.checksum(self.book.toc)

        yield Info('info', 'Checking metadata...')
        yield from self._check_meta()

        yield Info('info', 'Checking data files...')
        yield from self._check_data_dir()

        yield Info('info', 'Checking TOC...')
        yield from self._check_toc()
        yield from self._check_toc_empty_subtree()

        yield Info('info', 'Checking favicon cache...')
        yield from self._check_favicon_cache()

        # update files
        if self.book.checksum(self.book.meta) != book_meta_orig:
            yield Info('info', 'Saving changed meta files...')
            self.book.save_meta_files()

        if self.book.checksum(self.book.toc) != book_toc_orig:
            yield Info('info', 'Saving changed TOC files...')
            self.book.save_toc_files()

        yield Info('info',
                   f'Totally {self.cnt_items} items, {self.cnt_dirs} folders, '
                   f'{self.cnt_files} files, {util.format_filesize(self.cnt_bytes)}.')

        if self.resolve and (self.cnt_errors + self.cnt_warns > 0):
            yield Info('info',
                       f'Found {self.cnt_errors} errors, {self.cnt_warns} warnings. '
                       f'(Resolved {self.cnt_resolves})')
        else:
            yield Info('info', f'Found {self.cnt_errors} errors, {self.cnt_warns} warnings.')

    def _load_tree(self):
        try:
            self.book.load_meta_files()
        except TreeFileError as exc:
            raise RuntimeError(
                f'Malformed meta file '
                f'{self.book.get_subpath(exc.filename)!r}: {exc}'
            ) from exc
        except OSError as exc:
            raise RuntimeError(
                f'Failed to load meta file '
                f'{self.book.get_subpath(exc.filename)!r}: {exc.strerror}'
            ) from exc

        try:
            self.book.load_toc_files()
        except TreeFileError as exc:
            raise RuntimeError(
                f'Malformed TOC file '
                f'{self.book.get_subpath(exc.filename)!r}: {exc}'
            ) from exc
        except OSError as exc:
            raise RuntimeError(f'Failed to load TOC file {self.book.get_subpath(exc.filename)!r}: {exc.strerror}') from exc

    def _check_meta(self):
        items_invalid_id = {}
        items_missing_index = {}
        items_missing_index_file = {}
        items_missing_date = {}
        items_older_mtime = {}
        items_absolute_icon = {}

        for id, meta in self.book.meta.items():
            if meta is None:
                continue

            yield Info('debug', f'Checking item meta for {id!r}')

            # id
            if id in self.book.SPECIAL_ITEM_ID:
                yield Info('error', f'{id!r}: invalid ID (special item)')
                self.cnt_errors += 1
                items_invalid_id[id] = True

            # type
            type = meta.get('type', '')

            # index
            index = meta.get('index', '')
            index_file = None
            if index:
                file = os.path.join(self.book.data_dir, index)
                if os.path.isfile(file):
                    index_file = file
                    yield from self._check_index_file(id, index, file)
                else:
                    yield Info('error', f'{id!r}: missing index file {self.book.get_subpath(file)!r}')
                    self.cnt_errors += 1
                    items_missing_index_file[id] = True

                if type in {'folder', 'separator'}:
                    yield Info('warn', f'{id!r}: a {type!r} item should not have an index file.')
                    self.cnt_warns += 1
                elif type == 'bookmark' and not index.lower().endswith('.htm'):
                    yield Info('warn', f"{id!r}: a bookmark item should use '*.htm' as index file.")
                    self.cnt_warns += 1
            else:
                if type not in self.book.ITEM_TYPES_WITH_OPTIONAL_INDEX:
                    yield Info('error', f"{id!r}: missing 'index' property.")
                    self.cnt_errors += 1
                    items_missing_index[id] = True

            # icon
            icon = meta.get('icon')
            if icon:
                icon_parts = urlsplit(icon)
                if icon_parts.scheme:
                    yield Info('warn', f'{id!r}: icon {util.crop(icon, 256)!r} is an absolute URL')
                    self.cnt_warns += 1
                    items_absolute_icon[id] = True
                elif icon_parts.netloc:
                    yield Info('error', f'{id!r}: icon {util.crop(icon, 256)!r} is a protocol-relative URL')
                    self.cnt_errors += 1
                elif icon_parts.path.startswith('/'):
                    yield Info('error', f'{id!r}: icon {util.crop(icon, 256)!r} is a root-relative URL')
                    self.cnt_errors += 1
                else:
                    yield from self._check_favicon_file(id, meta)

            # create
            create = meta.get('create')
            if not create:
                yield Info('error', f"{id!r}: missing 'create' property.")
                self.cnt_errors += 1
                items_missing_date[id] = True

            # modify
            modify = meta.get('modify')
            if not modify:
                yield Info('error', f"{id!r}: missing 'modify' property.")
                self.cnt_errors += 1
                items_missing_date[id] = True
            elif index_file:
                try:
                    ts = os.stat(index_file).st_mtime
                except OSError:
                    pass
                else:
                    dt = datetime.fromtimestamp(ts - OLDER_MTIME_THRESHOLD)
                    mtime_check = util.datetime_to_id(dt)
                    if mtime_check > modify:
                        dt = datetime.fromtimestamp(ts)
                        mtime = util.datetime_to_id(dt)
                        yield Info('warn', f"{id!r}: 'modify' property ({modify}) is older than last modified time of index file ({mtime}).")
                        self.cnt_warns += 1
                        items_older_mtime[id] = True

        self.cnt_items += len(self.book.meta)

        if items_invalid_id and self.resolve_invalid_id:
            yield from self._resolve_invalid_id(items_invalid_id)

        if items_missing_index and self.resolve_missing_index:
            yield from self._resolve_missing_index(items_missing_index)

        if items_missing_index_file and self.resolve_missing_index_file:
            yield from self._resolve_missing_index_file(items_missing_index_file)

        if items_missing_date and self.resolve_missing_date:
            yield from self._resolve_missing_date(items_missing_date)

        if items_older_mtime and self.resolve_older_mtime:
            yield from self._resolve_older_mtime(items_older_mtime)

        if items_absolute_icon and self.resolve_absolute_icon:
            yield from self._resolve_absolute_icon(items_absolute_icon)

    def _check_index_file(self, id, index, file):
        pf = self._get_index_path_key(file)
        if pf in self.find_index_exclude:
            yield Info('warn', f'{id!r}: index file {self.book.get_subpath(file)!r} is used by another item.')
            self.cnt_warns += 1
        else:
            yield Info('debug', f'Excluding {pf!r} from index finding')
            self.find_index_exclude.add(pf)

            if index.endswith('/index.html'):
                pd = self._get_index_path_key(os.path.dirname(file))
                yield Info('debug', f'Excluding {pd!r} from index finding')
                self.find_index_exclude.add(pd)
            elif util.is_html(index):
                basename, ext = os.path.splitext(index)
                for suffix in SUPPORT_FOLDER_SUFFIXES:
                    p = self._get_index_path_key(os.path.join(os.path.dirname(file), f'{basename}{suffix}'))
                    yield Info('debug', f'Excluding {p!r} from index finding')
                    self.find_index_exclude.add(p)
            elif util.is_archive(index):
                try:
                    zh = zipfile.ZipFile(file)
                except zipfile.BadZipFile:
                    yield Info('error', f'{id!r}: corrupted archive file {self.book.get_subpath(file)!r}')
                    self.cnt_errors += 1
                    return

                with zh as zh:
                    if util.is_htz(index):
                        try:
                            zh.getinfo('index.html')
                        except KeyError:
                            yield Info('error', f"{id!r}: missing 'index.html' in archive file {self.book.get_subpath(file)!r}")
                            self.cnt_errors += 1
                    else:
                        if not util.get_maff_pages(zh):
                            yield Info('error', f'{id!r}: no valid page in archive file {self.book.get_subpath(file)!r}')
                            self.cnt_errors += 1

    def _check_favicon_file(self, id, meta):
        favicon_dir = os.path.join(self.book.tree_dir, 'favicon', '')
        file = self.book.get_icon_file(meta)
        file_ci = os.path.normcase(file)
        is_in_favicon_dir = file_ci.startswith(os.path.normcase(favicon_dir))

        if not os.path.isfile(file):
            yield Info('error', f'{id!r}: missing icon file {self.book.get_subpath(file)!r}')
            self.cnt_errors += 1
            return

        if is_in_favicon_dir:
            self.used_favicons.add(file_ci)

    def _check_toc(self):
        items_unreachable = {}
        items_missing_meta = {}
        ref_items_invalid = {}

        for id, ref_ids in self.book.toc.items():
            if ref_ids is None:
                continue

            yield Info('debug', f'Checking item TOC for {id!r}')

            # missing meta
            if not self.book.meta.get(id) and id not in self.book.SPECIAL_ITEM_ID:
                yield Info('error', f'{id!r}: invalid ID (missing metadata entry)')
                self.cnt_errors += 1
                items_missing_meta[id] = True

            # check referenced IDs
            for ref_id in ref_ids:
                self.seen_in_toc.add(ref_id)

                # special item ID is invalid
                if ref_id in self.book.SPECIAL_ITEM_ID:
                    yield Info('error', f'{id!r}: invalid reference ID {ref_id!r} (special item)')
                    self.cnt_errors += 1
                    ref_items_invalid.setdefault(id, {})[ref_id] = True
                    continue

                # missing meta
                if not self.book.meta.get(ref_id):
                    yield Info('error', f'{id!r}: invalid reference ID {ref_id!r} (missing metadata entry)')
                    self.cnt_errors += 1
                    ref_items_invalid.setdefault(id, {})[ref_id] = True
                    continue

        # items not reachable from TOC
        for id in self.book.meta:
            if id not in self.seen_in_toc and id not in self.book.SPECIAL_ITEM_ID:
                yield Info('error', f'{id!r}: not recheable from TOC.')
                self.cnt_errors += 1
                items_unreachable[id] = True

        if (items_missing_meta or ref_items_invalid) and self.resolve_toc_invalid:
            yield from self._resolve_toc_invalid(items_missing_meta, ref_items_invalid)

        if items_unreachable and self.resolve_toc_unreachable:
            yield from self._resolve_toc_unreachable(items_unreachable)

    def _check_toc_empty_subtree(self):
        # Calculate this after other TOC related issues are resolved,
        # as they might produce more empty lists.
        yield Info('debug', 'Checking empty lists in TOC...')

        items_empty_toc = {}
        for id, ref_ids in self.book.toc.items():
            if not ref_ids:
                yield Info('warn', f'{id!r}: TOC list is empty')
                self.cnt_warns += 1
                items_empty_toc[id] = True

        if items_empty_toc and self.resolve_toc_empty_subtree:
            yield from self._resolve_toc_empty_subtree(items_empty_toc)

    def _check_data_dir(self):
        unindexed_files = {}
        yield from self._check_data_dir_internal(self.book.data_dir, unindexed_files, find_index=True)

        if unindexed_files and self.resolve_unindexed_files:
            yield from self._resolve_unindexed_files(unindexed_files)

    def _check_data_dir_internal(self, data_dir, unindexed_files, find_index=True):
        yield Info('debug', f'Inspecting folder {self.book.get_subpath(data_dir)!r} (find_index={find_index!r})')

        if find_index:
            index = os.path.join(data_dir, 'index.html')
            if self._get_index_path_key(index) not in self.find_index_exclude and os.path.isfile(index):
                yield Info('warn', f'File {self.book.get_subpath(index)!r} not used as item index')
                self.cnt_warns += 1
                unindexed_files[index] = True

                yield Info('debug', f'Excluding {self.book.get_subpath(data_dir)!r} from index finding')
                find_index = False

        try:
            entries = os.scandir(data_dir)
        except FileNotFoundError:
            return
        except OSError as exc:
            yield Info('error', f'Failed to scan folder {self.book.get_subpath(exc.filename)!r}: {exc.strerror}', exc=exc)
            self.cnt_errors += 1
            return

        entries_to_handle = set()
        with entries as entries:
            for entry in entries:
                try:
                    assert not os.path.samefile(entry, self.wsb_dir)
                except AssertionError:
                    yield Info('debug', f'Skipped special {self.book.get_subpath(entry)!r}')
                    continue
                except OSError:
                    pass

                try:
                    assert not os.path.samefile(entry, self.book_wsb_dir)
                except AssertionError:
                    yield Info('debug', f'Skipped special {self.book.get_subpath(entry)!r}')
                    continue
                except OSError:
                    pass

                if entry.is_dir():
                    self.cnt_dirs += 1
                    entries_to_handle.add(entry)

                elif entry.is_file():
                    try:
                        self.cnt_bytes += entry.stat().st_size
                    except OSError as exc:
                        # e.g. a broken symlink can cause this
                        yield Info('error', f'Failed to access file {self.book.get_subpath(exc.filename)!r}: {exc.strerror}', exc=exc)
                        self.cnt_errors += 1
                    else:
                        self.cnt_files += 1

                    if not find_index:
                        continue

                    if self._get_index_path_key(entry) in self.find_index_exclude:
                        continue

                    basename, ext = os.path.splitext(entry.name)
                    if ext.lower() not in self.book.ITEM_INDEX_ALLOWED_EXT:
                        continue

                    entries_to_handle.add(entry)

                    if not util.is_html(entry.path):
                        continue

                    for suffix in SUPPORT_FOLDER_SUFFIXES:
                        p = self._get_index_path_key(os.path.join(data_dir, f'{basename}{suffix}'))
                        yield Info('debug', f'Excluding {p!r} from index finding')
                        self.find_index_exclude.add(p)

        for entry in sorted(entries_to_handle, key=lambda x: x.path):
            if entry.is_dir():
                p = self._get_index_path_key(entry)
                chk = find_index and p not in self.find_index_exclude
                yield from self._check_data_dir_internal(entry, unindexed_files, find_index=chk)

            elif entry.is_file():
                yield Info('warn', f'File {self.book.get_subpath(entry)!r} not used as item index')
                self.cnt_warns += 1
                unindexed_files[entry.path] = True

    def _check_favicon_cache(self):
        unused_icons = {}

        try:
            entries = os.scandir(os.path.join(self.book.tree_dir, 'favicon'))
        except FileNotFoundError:
            return
        except OSError as exc:
            yield Info('error', f'Failed to scan folder {self.book.get_subpath(exc.filename)!r}: {exc.strerror}', exc=exc)
            self.cnt_errors += 1
            return

        with entries as entries:
            for entry in entries:
                if os.path.normcase(entry.path) not in self.used_favicons:
                    yield Info('warn', f'Unused favicon file {self.book.get_subpath(entry)!r}.')
                    self.cnt_warns += 1
                    unused_icons[entry.path] = True

        if unused_icons and self.resolve_unused_icon:
            yield from self._resolve_unused_icon(unused_icons)

    def _get_index_path_key(self, path):
        return self.book.get_subpath(os.path.normcase(path))

    def _resolve_invalid_id(self, ids):
        yield Info('info', 'Removing items with invalid ID...')
        for id in ids:
            try:
                del self.book.meta[id]
            except KeyError:
                pass
            else:
                yield Info('info', f'Removed {id!r} from meta.')
                self.cnt_resolves += 1

    def _resolve_missing_index(self, ids):
        yield Info('info', 'Removing items missing index property...')
        for id in ids:
            try:
                del self.book.meta[id]
            except KeyError:
                pass
            else:
                yield Info('info', f'Removed {id!r} from meta.')
                self.cnt_resolves += 1

    def _resolve_missing_index_file(self, ids):
        yield Info('info', 'Removing items missing index file...')
        for id in ids:
            try:
                del self.book.meta[id]
            except KeyError:
                pass
            else:
                yield Info('info', f'Removed {id!r} from meta.')
                self.cnt_resolves += 1

    def _resolve_missing_date(self, ids):
        yield Info('info', 'Generating missing create/modify item property...')
        for id in ids:
            try:
                item = self.book.meta[id]
            except KeyError:
                continue

            if not item.get('create'):
                create = generate_item_create(self.book, id)
                if create:
                    item['create'] = create
                    yield Info('info', f"Added 'create' property for {id!r}.")
                    self.cnt_resolves += 1

            if not item.get('modify'):
                modify = generate_item_modify(self.book, id)
                if modify:
                    item['modify'] = modify
                    yield Info('info', f"Added 'modify' property for {id!r}.")
                    self.cnt_resolves += 1

    def _resolve_older_mtime(self, ids):
        yield Info('info', 'Updating items with older modify property...')
        for id in ids:
            index = self.book.meta[id].get('index')
            if index:
                file = os.path.join(self.book.data_dir, index)
                try:
                    ts = os.stat(file).st_mtime
                except OSError:
                    pass
                else:
                    dt = datetime.fromtimestamp(ts)
                    mtime = util.datetime_to_id(dt)
                    if mtime > self.book.meta[id].get('modify'):
                        self.book.meta[id]['modify'] = mtime
                        yield Info('info', f"Updated 'modify' property for {id!r}.")
                        self.cnt_resolves += 1

    def _resolve_toc_invalid(self, items_missing_meta, ref_items_invalid):
        yield Info('info', 'Removing invalid items from TOC...')

        for id in items_missing_meta:
            try:
                del self.book.toc[id]
            except KeyError:
                pass
            else:
                yield Info('info', f'Removed {id!r} from TOC.')
                self.cnt_resolves += 1

        for id, ref_ids in ref_items_invalid.items():
            for ref_id in ref_ids:
                try:
                    self.book.toc[id].remove(ref_id)
                except (KeyError, ValueError):
                    pass
                else:
                    yield Info('info', f'Removed {ref_id!r} from the subtree of {id!r}.')
                    self.cnt_resolves += 1

    def _resolve_toc_unreachable(self, ids):
        yield Info('info', 'Adding unreachable items to root TOC...')
        toc = self.book.toc.setdefault(self.book.ROOT_ITEM_ID, [])
        for id in ids:
            toc.append(id)
            yield Info('info', f'Added {id!r} to root TOC.')
            self.cnt_resolves += 1

    def _resolve_toc_empty_subtree(self, ids):
        yield Info('info', 'Removing empty item lists from TOC...')
        for id in ids:
            try:
                del self.book.toc[id]
            except KeyError:
                pass
            else:
                yield Info('info', f'Removed {id!r} from TOC.')
                self.cnt_resolves += 1

    def _resolve_unindexed_files(self, files):
        yield Info('info', 'Indexing unindexed files...')
        indexer = Indexer(self.book)
        indexed = yield from indexer.run(files)

        favicon_dir = os.path.join(self.book.tree_dir, 'favicon', '')

        for id in indexed:
            # add to TOC if not seen
            if id not in self.seen_in_toc:
                root = self.book.ROOT_ITEM_ID
                target_index = 0 if self.book.config['new_at_top'] else len(self.book.toc.get(root, ()))
                self.book.toc.setdefault(root, []).insert(target_index, id)
                self.seen_in_toc.add(id)

            # record added cached favicons
            index = self.book.meta[id].get('index')
            icon = self.book.meta[id].get('icon')
            if icon:
                file = os.path.normpath(os.path.join(self.book.data_dir, os.path.dirname(index), unquote(icon)))
                file_ci = os.path.normcase(file)
                is_in_favicon_dir = file_ci.startswith(os.path.normcase(favicon_dir))
                if is_in_favicon_dir:
                    self.used_favicons.add(file_ci)

            self.cnt_resolves += 1

    def _resolve_absolute_icon(self, ids):
        yield Info('info', 'Caching favicons with absolute URL...')
        generator = FavIconCacher(self.book)
        cached = yield from generator.run(ids)

        for _id, file in cached.items():
            self.used_favicons.add(os.path.normcase(file))
            self.cnt_resolves += 1

    def _resolve_unused_icon(self, files):
        yield Info('info', 'Removing unused favicons...')
        for file in files:
            try:
                self.book.backup(file)
                os.remove(file)
            except FileNotFoundError:
                pass
            else:
                yield Info('info', f'Removed {self.book.get_subpath(file)!r}.')
                self.cnt_resolves += 1


def run(host, book_ids=None, *, lock=True, backup=True, **kwargs):
    start = time.time()

    if isinstance(host, Host):
        pass
    elif isinstance(host, str):
        host = Host(host)
    else:
        host = Host(*host)

    if backup:
        host.init_auto_backup(note='check')
        yield Info('info', f'Prepared backup at {host.get_subpath(host._auto_backup_dir)!r}.')
        yield Info('info', '----------------------------------------------------------------------')

    try:
        # handle all books if none specified
        first = True
        for book_id in book_ids or host.books:
            if first:
                first = False
            else:
                yield Info('info', '----------------------------------------------------------------------')

            try:
                book = host.books[book_id]
            except KeyError:
                # skip invalid book ID
                yield Info('warn', f'Skipped invalid book {book_id!r}.')
                continue

            yield Info('debug', f'Loading book {book_id!r}...')

            if book.no_tree:
                yield Info('info', f'Skipped book {book_id!r} ({book.name!r}) (no_tree).')
                continue

            yield Info('info', f'Checking book {book_id!r} ({book.name!r}).')
            lh = book.get_tree_lock(persist=lock).acquire() if lock else nullcontext()
            with lh:
                generator = BookChecker(book, **kwargs)
                yield from generator.run()

            yield Info('info', 'Done.')
    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
        return
    finally:
        if backup:
            host.init_auto_backup(False)

    yield Info('info', '----------------------------------------------------------------------')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



"""Miscellaneous utilities
"""
import binascii
import codecs
import hashlib
import importlib
import math
import os
import re
import sys
from base64 import b64decode
from collections import namedtuple
from contextlib import nullcontext
from datetime import datetime, timezone
from ipaddress import AddressValueError, IPv6Address
from urllib.parse import quote, unquote_to_bytes, urljoin, urlsplit
from urllib.request import pathname2url, url2pathname

import lxml.html
from lxml import etree

from .._polyfill import mimetypes, zipfile

#########################################################################
# Common classes and objects handling
#########################################################################

# common namedtuple for yielded messages for certain classes
Info = namedtuple('Info', ('type', 'msg', 'data', 'exc'), defaults=(None, None))


def import_module_file(ns, file):
    try:
        return sys.modules[ns]
    except KeyError:
        pass

    spec = importlib.util.spec_from_file_location(ns, file)
    module = importlib.util.module_from_spec(spec)
    sys.modules[ns] = module
    try:
        spec.loader.exec_module(module)
    except Exception:
        del sys.modules[ns]
        raise
    return module


#########################################################################
# ScrapBook related path/file/string/etc handling
#########################################################################

REGEX_ID_TO_DATETIME = re.compile(r'^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})(\d{3})$')
REGEX_ID_TO_DATETIME_LEGACY = re.compile(r'^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})(\d{0,9})$')


def datetime_to_id(t=None):
    """Convert a datetime to webscrapbook ID.

    Args:
        t: datetime. Create an ID for now if None.
    """
    if t is None:
        t = datetime.now(timezone.utc)
    else:
        # convert to UTC datetime
        t = t.astimezone(timezone.utc)

    return (f'{t.year}{t.month:02}{t.day:02}{t.hour:02}{t.minute:02}'
            f'{t.second:02}{int(t.microsecond * 0.001):03}')


def id_to_datetime(id):
    """Convert a webscrapbook ID to datetime.
    """
    m = REGEX_ID_TO_DATETIME.search(id)
    if m:
        try:
            return datetime(
                int(m.group(1)),
                int(m.group(2)),
                int(m.group(3)),
                int(m.group(4)),
                int(m.group(5)),
                int(m.group(6)),
                int(m.group(7)) * 1000,
                timezone.utc,
            )
        except ValueError:
            pass
    return None


def datetime_to_id_legacy(t=None):
    """Convert a datetime to legacy ScrapBook ID.

    Args:
        t: datetime. Create an ID for now if None.
    """
    if t is None:
        t = datetime.now()
    else:
        # convert to local datetime
        t = t.astimezone()

    return f'{t.year}{t.month:02}{t.day:02}{t.hour:02}{t.minute:02}{t.second:02}'


def id_to_datetime_legacy(id):
    """Convert a legacy ScrapBook ID to datetime.
    """
    m = REGEX_ID_TO_DATETIME_LEGACY.search(id)
    if m:
        try:
            ms = m.group(7)
            try:
                ms = int(ms) * 10 ** (6 - len(ms))
            except ValueError:
                ms = 0

            return datetime(
                int(m.group(1)),
                int(m.group(2)),
                int(m.group(3)),
                int(m.group(4)),
                int(m.group(5)),
                int(m.group(6)),
                int(ms),
            )
        except ValueError:
            pass
    return None


def unify_pathsep(text):
    """Convert non-standard path separators to '/'."""
    if os.sep != '/':
        return text.replace(os.sep, '/')
    return text


def validate_filename(filename, force_ascii=False):
    """Transliterates the given string to be a safe filename

    See also: scrapbook.validateFilename of WebScrapBook.
    """
    fn = filename

    # common restrictions
    # - collapse document spaces
    fn = re.sub(r'[\t\n\f\r]+', ' ', fn)

    # - control chars are bad for filename
    fn = re.sub(r'[\x00-\x1F\x7F\x80-\x9F]+', '', fn)

    # - bad chars on most OS
    fn = re.sub(r'[:"?*\\/|<>]', '_', fn)

    # Windows restrictions
    # - leading/trailing spaces and dots
    fn = re.sub(r'^ +', '', fn)
    fn = re.sub(r'[. ]+$', '', fn)
    fn = re.sub(r'^\.', '_.', fn)

    # - reserved filenames
    fn = re.sub(r'^(CON|PRN|AUX|NUL|COM\d|LPT\d)((?:\..+)?)$', r'\g<1>_\g<2>', fn, flags=re.I)

    if force_ascii:
        fn = quote(fn, safe="""!_#$%&'()*+,-./:;<=>?@[\\]^_`{|}~""")

    # prevent empty filename
    fn = fn or '_'

    return fn


#########################################################################
# String handling
#########################################################################

def cropped(text, width=70, ellipsis='...'):
    if len(text) > width:
        return (text[:max(width - len(ellipsis), 0)], ellipsis)
    return (text, '')


def crop(text, width=70, ellipsis='...'):
    return ''.join(cropped(text, width, ellipsis))


REGEX_FORMAT_STRING = re.compile(r'%(\w*)%')


def format_string(text, mapping):
    """A very simple implementation for string formatting with placeholders.

    - Only single special char '%' is used, making it easy to implement for
      both Python and JavaScript.
    - Special char itself can be easily escaped.
    - Does not raise an exception if not formatted correctly.
    - Good for user-provided strings.
    """
    def formatter(m):
        k = m.group(1)
        if k == '':
            return '%'
        return mapping.get(k, '')

    return REGEX_FORMAT_STRING.sub(formatter, text)


REGEX_COMPRESS_CODE = re.compile(r'[^\S　]+')


def compress_code(code):
    return REGEX_COMPRESS_CODE.sub(' ', code)


#########################################################################
# Codecs and text encoding
#########################################################################

# ref: https://encoding.spec.whatwg.org/#names-and-labels
LABEL_ENCODING_MAPPING = {
    'unicode-1-1-utf-8': 'UTF-8',
    'unicode11utf8': 'UTF-8',
    'unicode20utf8': 'UTF-8',
    'utf-8': 'UTF-8',
    'utf8': 'UTF-8',
    'x-unicode20utf8': 'UTF-8',
    '866': 'IBM866',
    'cp866': 'IBM866',
    'csibm866': 'IBM866',
    'ibm866': 'IBM866',
    'csisolatin2': 'ISO-8859-2',
    'iso-8859-2': 'ISO-8859-2',
    'iso-ir-101': 'ISO-8859-2',
    'iso8859-2': 'ISO-8859-2',
    'iso88592': 'ISO-8859-2',
    'iso_8859-2': 'ISO-8859-2',
    'iso_8859-2:1987': 'ISO-8859-2',
    'l2': 'ISO-8859-2',
    'latin2': 'ISO-8859-2',
    'csisolatin3': 'ISO-8859-3',
    'iso-8859-3': 'ISO-8859-3',
    'iso-ir-109': 'ISO-8859-3',
    'iso8859-3': 'ISO-8859-3',
    'iso88593': 'ISO-8859-3',
    'iso_8859-3': 'ISO-8859-3',
    'iso_8859-3:1988': 'ISO-8859-3',
    'l3': 'ISO-8859-3',
    'latin3': 'ISO-8859-3',
    'csisolatin4': 'ISO-8859-4',
    'iso-8859-4': 'ISO-8859-4',
    'iso-ir-110': 'ISO-8859-4',
    'iso8859-4': 'ISO-8859-4',
    'iso88594': 'ISO-8859-4',
    'iso_8859-4': 'ISO-8859-4',
    'iso_8859-4:1988': 'ISO-8859-4',
    'l4': 'ISO-8859-4',
    'latin4': 'ISO-8859-4',
    'csisolatincyrillic': 'ISO-8859-5',
    'cyrillic': 'ISO-8859-5',
    'iso-8859-5': 'ISO-8859-5',
    'iso-ir-144': 'ISO-8859-5',
    'iso8859-5': 'ISO-8859-5',
    'iso88595': 'ISO-8859-5',
    'iso_8859-5': 'ISO-8859-5',
    'iso_8859-5:1988': 'ISO-8859-5',
    'arabic': 'ISO-8859-6',
    'asmo-708': 'ISO-8859-6',
    'csiso88596e': 'ISO-8859-6',
    'csiso88596i': 'ISO-8859-6',
    'csisolatinarabic': 'ISO-8859-6',
    'ecma-114': 'ISO-8859-6',
    'iso-8859-6': 'ISO-8859-6',
    'iso-8859-6-e': 'ISO-8859-6',
    'iso-8859-6-i': 'ISO-8859-6',
    'iso-ir-127': 'ISO-8859-6',
    'iso8859-6': 'ISO-8859-6',
    'iso88596': 'ISO-8859-6',
    'iso_8859-6': 'ISO-8859-6',
    'iso_8859-6:1987': 'ISO-8859-6',
    'csisolatingreek': 'ISO-8859-7',
    'ecma-118': 'ISO-8859-7',
    'elot_928': 'ISO-8859-7',
    'greek': 'ISO-8859-7',
    'greek8': 'ISO-8859-7',
    'iso-8859-7': 'ISO-8859-7',
    'iso-ir-126': 'ISO-8859-7',
    'iso8859-7': 'ISO-8859-7',
    'iso88597': 'ISO-8859-7',
    'iso_8859-7': 'ISO-8859-7',
    'iso_8859-7:1987': 'ISO-8859-7',
    'sun_eu_greek': 'ISO-8859-7',
    'csiso88598e': 'ISO-8859-8',
    'csisolatinhebrew': 'ISO-8859-8',
    'hebrew': 'ISO-8859-8',
    'iso-8859-8': 'ISO-8859-8',
    'iso-8859-8-e': 'ISO-8859-8',
    'iso-ir-138': 'ISO-8859-8',
    'iso8859-8': 'ISO-8859-8',
    'iso88598': 'ISO-8859-8',
    'iso_8859-8': 'ISO-8859-8',
    'iso_8859-8:1988': 'ISO-8859-8',
    'visual': 'ISO-8859-8',
    'csiso88598i': 'ISO-8859-8-I',
    'iso-8859-8-i': 'ISO-8859-8-I',
    'logical': 'ISO-8859-8-I',
    'csisolatin6': 'ISO-8859-10',
    'iso-8859-10': 'ISO-8859-10',
    'iso-ir-157': 'ISO-8859-10',
    'iso8859-10': 'ISO-8859-10',
    'iso885910': 'ISO-8859-10',
    'l6': 'ISO-8859-10',
    'latin6': 'ISO-8859-10',
    'iso-8859-13': 'ISO-8859-13',
    'iso8859-13': 'ISO-8859-13',
    'iso885913': 'ISO-8859-13',
    'iso-8859-14': 'ISO-8859-14',
    'iso8859-14': 'ISO-8859-14',
    'iso885914': 'ISO-8859-14',
    'csisolatin9': 'ISO-8859-15',
    'iso-8859-15': 'ISO-8859-15',
    'iso8859-15': 'ISO-8859-15',
    'iso885915': 'ISO-8859-15',
    'iso_8859-15': 'ISO-8859-15',
    'l9': 'ISO-8859-15',
    'iso-8859-16': 'ISO-8859-16',
    'cskoi8r': 'KOI8-R',
    'koi': 'KOI8-R',
    'koi8': 'KOI8-R',
    'koi8-r': 'KOI8-R',
    'koi8_r': 'KOI8-R',
    'koi8-ru': 'KOI8-U',
    'koi8-u': 'KOI8-U',
    'csmacintosh': 'macintosh',
    'mac': 'macintosh',
    'macintosh': 'macintosh',
    'x-mac-roman': 'macintosh',
    'dos-874': 'windows-874',
    'iso-8859-11': 'windows-874',
    'iso8859-11': 'windows-874',
    'iso885911': 'windows-874',
    'tis-620': 'windows-874',
    'windows-874': 'windows-874',
    'cp1250': 'windows-1250',
    'windows-1250': 'windows-1250',
    'x-cp1250': 'windows-1250',
    'cp1251': 'windows-1251',
    'windows-1251': 'windows-1251',
    'x-cp1251': 'windows-1251',
    'ansi_x3.4-1968': 'windows-1252',
    'ascii': 'windows-1252',
    'cp1252': 'windows-1252',
    'cp819': 'windows-1252',
    'csisolatin1': 'windows-1252',
    'ibm819': 'windows-1252',
    'iso-8859-1': 'windows-1252',
    'iso-ir-100': 'windows-1252',
    'iso8859-1': 'windows-1252',
    'iso88591': 'windows-1252',
    'iso_8859-1': 'windows-1252',
    'iso_8859-1:1987': 'windows-1252',
    'l1': 'windows-1252',
    'latin1': 'windows-1252',
    'us-ascii': 'windows-1252',
    'windows-1252': 'windows-1252',
    'x-cp1252': 'windows-1252',
    'cp1253': 'windows-1253',
    'windows-1253': 'windows-1253',
    'x-cp1253': 'windows-1253',
    'cp1254': 'windows-1254',
    'csisolatin5': 'windows-1254',
    'iso-8859-9': 'windows-1254',
    'iso-ir-148': 'windows-1254',
    'iso8859-9': 'windows-1254',
    'iso88599': 'windows-1254',
    'iso_8859-9': 'windows-1254',
    'iso_8859-9:1989': 'windows-1254',
    'l5': 'windows-1254',
    'latin5': 'windows-1254',
    'windows-1254': 'windows-1254',
    'x-cp1254': 'windows-1254',
    'cp1255': 'windows-1255',
    'windows-1255': 'windows-1255',
    'x-cp1255': 'windows-1255',
    'cp1256': 'windows-1256',
    'windows-1256': 'windows-1256',
    'x-cp1256': 'windows-1256',
    'cp1257': 'windows-1257',
    'windows-1257': 'windows-1257',
    'x-cp1257': 'windows-1257',
    'cp1258': 'windows-1258',
    'windows-1258': 'windows-1258',
    'x-cp1258': 'windows-1258',
    'x-mac-cyrillic': 'x-mac-cyrillic',
    'x-mac-ukrainian': 'x-mac-cyrillic',
    'chinese': 'GBK',
    'csgb2312': 'GBK',
    'csiso58gb231280': 'GBK',
    'gb2312': 'GBK',
    'gb_2312': 'GBK',
    'gb_2312-80': 'GBK',
    'gbk': 'GBK',
    'iso-ir-58': 'GBK',
    'x-gbk': 'GBK',
    'gb18030': 'gb18030',
    'big5': 'Big5',
    'big5-hkscs': 'Big5',
    'cn-big5': 'Big5',
    'csbig5': 'Big5',
    'x-x-big5': 'Big5',
    'cseucpkdfmtjapanese': 'EUC-JP',
    'euc-jp': 'EUC-JP',
    'x-euc-jp': 'EUC-JP',
    'csiso2022jp': 'ISO-2022-JP',
    'iso-2022-jp': 'ISO-2022-JP',
    'csshiftjis': 'Shift_JIS',
    'ms932': 'Shift_JIS',
    'ms_kanji': 'Shift_JIS',
    'shift-jis': 'Shift_JIS',
    'shift_jis': 'Shift_JIS',
    'sjis': 'Shift_JIS',
    'windows-31j': 'Shift_JIS',
    'x-sjis': 'Shift_JIS',
    'cseuckr': 'EUC-KR',
    'csksc56011987': 'EUC-KR',
    'euc-kr': 'EUC-KR',
    'iso-ir-149': 'EUC-KR',
    'korean': 'EUC-KR',
    'ks_c_5601-1987': 'EUC-KR',
    'ks_c_5601-1989': 'EUC-KR',
    'ksc5601': 'EUC-KR',
    'ksc_5601': 'EUC-KR',
    'windows-949': 'EUC-KR',
    'unicodefffe': 'UTF-16BE',
    'utf-16be': 'UTF-16BE',
    'csunicode': 'UTF-16LE',
    'iso-10646-ucs-2': 'UTF-16LE',
    'ucs-2': 'UTF-16LE',
    'unicode': 'UTF-16LE',
    'unicodefeff': 'UTF-16LE',
    'utf-16': 'UTF-16LE',
    'utf-16le': 'UTF-16LE',
    'csiso2022kr': 'replacement',
    'hz-gb-2312': 'replacement',
    'iso-2022-cn': 'replacement',
    'iso-2022-cn-ext': 'replacement',
    'iso-2022-kr': 'replacement',
    'replacement': 'replacement',
    'x-user-defined': 'x-user-defined',
}

# all lower-case
CODECS_MAPPING = {
    'big5': 'big5hkscs',
    'x-mac-cyrillic': 'mac_cyrillic',
    'replacement': None,
    'x-user-defined': None,
}


def fix_codec(name):
    """Remap codec name

    Map a browser-supported encoding label to the corresponding encoding.

    Some codecs are widely used and de-facto standard for web browsers. For
    example, most browsers display cp950 extended chars correctly even if the
    charset of the web page is defined as 'big5'. To prevent unexpected
    gibberish when we try to parse text in Python, we need to remap the codec
    name of a web page from 'big5' to 'big5hkscs' using this first.
    """
    try:
        name = LABEL_ENCODING_MAPPING[name.lower()]
    except KeyError:
        pass

    try:
        return CODECS_MAPPING[name.lower()]
    except KeyError:
        return name


# starting of BOM32 is equal to BOM16, so check the former first
BOM_DETECTORS = [
    ('UTF-8-SIG', codecs.BOM_UTF8),
    ('UTF-32-LE', codecs.BOM_UTF32_LE),
    ('UTF-32-BE', codecs.BOM_UTF32_BE),
    ('UTF-16-LE', codecs.BOM_UTF16_LE),
    ('UTF-16-BE', codecs.BOM_UTF16_BE),
]


def sniff_bom(fh):
    """Sniff a possibly existing BOM

    Args:
        fh: an opened file handler, must be seekable.

    Return:
        str: corresponding codec name for a found BOM if a BOM is found (and
            sets pointer at the position after the BOM), or None otherwise.
    """
    # will read less if the file is smaller
    raw = fh.read(4)

    for enc, bom in BOM_DETECTORS:
        if raw.startswith(bom):
            fh.seek(len(bom))
            return enc

    fh.seek(0)
    return None


#########################################################################
# URL and string
#########################################################################

def is_nullhost(host):
    """Determine if given host is 0.0.0.0 equivalent.
    """
    if host == '0.0.0.0':
        return True

    try:
        if IPv6Address(host) == IPv6Address('::'):
            return True
    except AddressValueError:
        pass

    return False


def is_localhost(host):
    """Determine if given host is localhost equivalent.
    """
    if host in ('localhost', '127.0.0.1'):
        return True

    try:
        if IPv6Address(host) == IPv6Address('::1'):
            return True
    except AddressValueError:
        pass

    return False


def get_relative_url(path, start, path_is_dir=True, start_is_dir=True):
    """Get a relative URL (quoted) from filesystem start to path
    """
    if not start_is_dir:
        start = os.path.dirname(start)
    rel_path = os.path.relpath(path, start)
    if path_is_dir:
        rel_path = os.path.join(rel_path, '')
    return pathname2url(rel_path)  # this quotes URL


#########################################################################
# Filesystem related manipulation
#########################################################################

def checksum(file, method='sha1', chunk_size=4096):
    """Calculate the checksum of a file.

    Args:
        file: str, path-like, or file-like bytes object
    """
    try:
        fh = open(file, 'rb')
    except TypeError:
        fh = nullcontext(file)

    with fh as fh:
        h = hashlib.new(method)
        while True:
            chunk = fh.read(chunk_size)
            if not chunk:
                break
            h.update(chunk)

        return h.hexdigest()


def format_filesize(bytes, si=False, space=' '):
    """Convert file size from bytes to human readable presentation.
    """
    try:
        bytes = int(bytes)
    except (ValueError, TypeError):
        return ''

    if si:
        thresh = 1000
        units = ['B', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']
    else:
        thresh = 1024
        units = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']

    e = math.floor(math.log(max(1, bytes)) / math.log(thresh))
    e = min(e, len(units) - 1)
    n = bytes / thresh ** e
    tpl = '{:.1f}{}{}' if (e >= 1 and n < 10) else '{:.0f}{}{}'  # noqa: P103
    return tpl.format(n, space, units[e])


COMPRESSIBLE_TYPES = {
    'application/xml',

    # historical non-text/* javascript types
    # ref: https://mimesniff.spec.whatwg.org/
    'application/javascript',
    'application/ecmascript',
    'application/x-ecmascript',
    'application/x-javascript',

    'application/json',
}

COMPRESSIBLE_SUFFIXES = {
    '+xml',
    '+json',
}


def is_compressible(mimetype):
    """Guess if the given mimetype is compressible."""
    if not mimetype:
        return False

    if mimetype.startswith('text/'):
        return True

    if mimetype in COMPRESSIBLE_TYPES:
        return True

    for suffix in COMPRESSIBLE_SUFFIXES:
        if mimetype.endswith(suffix):
            return True

    return False


def mime_is_html(mime):
    return mime in {'text/html', 'application/xhtml+xml'}


def mime_is_xhtml(mime):
    return mime == 'application/xhtml+xml'


def mime_is_svg(mime):
    return mime == 'image/svg+xml'


def mime_is_archive(mime):
    return mime in {'application/html+zip', 'application/x-maff'}


def mime_is_htz(mime):
    return mime == 'application/html+zip'


def mime_is_maff(mime):
    return mime == 'application/x-maff'


def mime_is_markdown(mime):
    return mime in {'text/markdown'}


def mime_is_wsba(mime):
    return mime in {'application/wsba+zip'}


def is_html(filename):
    mime, _ = mimetypes.guess_type(filename)
    return mime_is_html(mime)


def is_xhtml(filename):
    mime, _ = mimetypes.guess_type(filename)
    return mime_is_xhtml(mime)


def is_svg(filename):
    mime, _ = mimetypes.guess_type(filename)
    return mime_is_svg(mime)


def is_archive(filename):
    mime, _ = mimetypes.guess_type(filename)
    return mime_is_archive(mime)


def is_htz(filename):
    mime, _ = mimetypes.guess_type(filename)
    return mime_is_htz(mime)


def is_maff(filename):
    mime, _ = mimetypes.guess_type(filename)
    return mime_is_maff(mime)


def is_markdown(filename):
    mime, _ = mimetypes.guess_type(filename)
    return mime_is_markdown(mime)


def is_wsba(filename):
    mime, _ = mimetypes.guess_type(filename)
    return mime_is_wsba(mime)


#########################################################################
# HTTP manipulation
#########################################################################

HEADER_OWS = r'[\t ]*'
HEADER_TOKEN = r"[!#$%&'*+.0-9A-Z^_`a-z|~-]+"
HEADER_QUOTED_STRING = r'(?:"[^"]*(?:\.[^"]*)*")'


ContentType = namedtuple('ContentType', ('type', 'parameters'))

CONTENT_TYPE_REGEX = re.compile(fr'^{HEADER_TOKEN}/{HEADER_TOKEN}')
CONTENT_TYPE_REGEX_PARAMETER = re.compile(fr"""
    ^
    {HEADER_OWS}
    ;
    {HEADER_OWS}
    ({HEADER_TOKEN})
    =
    ([^\t ;"]*(?:{HEADER_QUOTED_STRING}[^\t ;"]*)*)
    """, re.X)


def parse_content_type(string):
    """Parse content type header.

    Return:
        ContentType: type and parameter keys are all lower case.
    """
    type = None
    parameters = {}

    if not string:
        return ContentType(type, parameters)

    m = CONTENT_TYPE_REGEX.search(string)
    if m:
        string = string[m.end():]
        type = m.group(0).lower()

        while True:
            m = CONTENT_TYPE_REGEX_PARAMETER.search(string)
            if not m:
                break

            string = string[m.end():]
            field = m.group(1).lower()
            value = m.group(2)

            if value.startswith('"'):
                # any valid value with leading '"' must be ".*"
                value = value[1:-1]

            parameters[field] = value

    return ContentType(type, parameters)


DataUri = namedtuple('DataUri', ('bytes', 'mime', 'parameters'))

PARSE_DATAURI_REGEX_FIELDS = re.compile(r'^data:([^,]*?)(;base64)?,([^#]*)', re.I)
PARSE_DATAURI_REGEX_KEY_VALUE = re.compile(r'^(.*?)=(.*?)$')


class DataUriMalformedError(Exception):
    pass


def parse_datauri(datauri):
    """Parse a Data URI

    Args:
        datauri: the data URI string

    Returns:
        DataUri: a tuple containing information

    Raises:
        DataUriMalformedError
    """
    match_fields = PARSE_DATAURI_REGEX_FIELDS.search(datauri)
    if not match_fields:
        raise DataUriMalformedError('Malformed fields')

    mediatype = match_fields.group(1)
    base64 = bool(match_fields.group(2))
    data = match_fields.group(3)

    parts = mediatype.split(';')
    mime = parts.pop(0)
    parameters = {}
    for part in parts:
        match_key_value = PARSE_DATAURI_REGEX_KEY_VALUE.search(part)
        if match_key_value:
            parameters[match_key_value.group(1).lower()] = match_key_value.group(2)

    if base64:
        try:
            bytes_ = b64decode(data)
        except binascii.Error as exc:
            raise DataUriMalformedError(f'Malformed base64 sequence: {exc}') from exc
    else:
        # decode precent-encoding to corresponding byte
        # non-ASCII chars are encoded as UTF-8 bytes
        bytes_ = unquote_to_bytes(data)

    return DataUri(bytes_, mime, parameters)


#########################################################################
# HTML manipulation
#########################################################################

def _get_html_charset(fh, quickly=True):
    """Internal method to read a charset from meta charset.
    """
    try:
        for _event, elem in etree.iterparse(fh, encoding='ISO-8859-1', html=True, events=('start',), tag=('meta', 'body')):
            if elem.tag == 'meta':
                charset = elem.attrib.get('charset')
                if charset:
                    return charset.strip()

                if elem.attrib.get('http-equiv', '').lower() == 'content-type':
                    _, params = parse_content_type(elem.attrib.get('content', ''))
                    charset = params.get('charset')
                    if charset:
                        return charset

            elif elem.tag == 'body':
                # presume that no <meta> will appear after <body> start
                # for a normal HTML to exit early
                if quickly:
                    return None

            # clean up to save memory
            elem.clear()
            while elem.getprevious() is not None:
                try:
                    del elem.getparent()[0]
                except TypeError:
                    # broken html may generate extra root elem
                    break
    except etree.Error:
        pass

    return None


def get_html_charset(file, default='UTF-8', none_from_bom=True, quickly=True):
    """Search for the correct charset to read an HTML file.

    Args:
        file: str, path-like, or file-like bytes object
        default: fallback encoding if not found
        none_from_bom: True to return None if charset is determined from BOM
            (to prevent error for lxml)
        quickly: True to exit early for normal HTML files
    """
    try:
        fh = open(file, 'rb')
    except TypeError:
        fh = nullcontext(file)
    except FileNotFoundError:
        fh = None

    if fh:
        with fh as fh:
            # Seek for the correct charset (encoding).
            # If a charset is not specified, lxml may select a wrong encoding for
            # the entire document if there is text before first meta charset.
            charset = sniff_bom(fh)
            if charset:
                if none_from_bom:
                    # lxml does not accept "UTF-16-LE" or so, but can auto-detect
                    # encoding from BOM if encoding is None
                    # ref: https://bugs.launchpad.net/lxml/+bug/1463610
                    return None

                return charset

            charset = _get_html_charset(fh, quickly=quickly)

            if charset is None:
                charset = default

            if charset is not None:
                charset = fix_codec(charset)

            return charset

    return default


def load_html_tree(file, options=None):
    """Load HTML document tree.

    Args:
        file: str, path-like, or file-like bytes object
        options: additional options for the HTML parser
    """
    try:
        fh = open(file, 'rb')
    except TypeError:
        fh = nullcontext(file)
    except FileNotFoundError:
        fh = None

    if not fh:
        return None

    with fh as fh:
        charset = get_html_charset(fh)
        fh.seek(0)

        try:
            return lxml.html.parse(fh, lxml.html.HTMLParser(encoding=charset, **(options or {})))
        except etree.Error:
            return None


MetaRefreshInfo = namedtuple('MetaRefreshInfo', ('time', 'target', 'context'))

# ref: https://html.spec.whatwg.org/multipage/semantics.html#attr-meta-http-equiv-refresh
META_REFRESH_REGEX = re.compile(r"""
    ^
    [\t\n\f\r ]*
    (?P<time>\d+)
    (?:\.[\d.]*)?
    (?:
        (?=[\t\n\f\r ;,])
        [\t\n\f\r ]*
        [;,]?
        [\t\n\f\r ]*
        (?:url[\t\n\f\r ]*=[\t\n\f\r ]*)?
        (?P<target>.*)
    )?
    $
    """, re.I + re.X)

# meta refresh in these tags does not always work
META_REFRESH_CONTEXT_TAGS = {
    'title',
    # 'style', 'script',  # not visible by lxml
    # 'frame',  # self-closing tag
    'iframe',
    # 'object', 'applet',  # refresh works in the browser
    # 'audio', 'video',  # refresh works in the browser
    # 'canvas',  # refresh works in the browser
    'noframes', 'noscript', 'noembed',
    'textarea',
    'template',
    # 'svg', 'math',  # refresh works in the browser
    'xmp',
    # 'parsererror',  # doesn't appear in lxml for xhtml
}

# meta refresh in these tags should never work
META_REFRESH_FORBID_TAGS = {
    'title',
    'textarea',
    'template',
    'xmp',
}


class MetaRefreshError(Exception):
    pass


class MetaRefreshCircularError(MetaRefreshError):
    pass


def parse_meta_refresh_content(string, contexts=None):
    """Parse a HTTP header like meta refresh content.
    """
    m = META_REFRESH_REGEX.search(string)
    if not m:
        return MetaRefreshInfo(time=None, target=None, context=None)

    try:
        time = int(m.group('time'))
    except ValueError:
        time = 0

    target = m.group('target')
    if target is not None:
        for qchar in ('"', "'"):
            if target.startswith(qchar):
                try:
                    pos = target.index(qchar, 1)
                except ValueError:
                    pos = None
                target = target[1:pos]
                break
        target = target.strip('\t\n\f\r ')

    context = contexts.copy() if contexts else None

    return MetaRefreshInfo(time=time, target=target or '', context=context)


def iter_meta_refresh(file, encoding=None):
    """Iterate through meta refreshes from a file.

    Args:
        file: str, path-like, or file-like bytes object
        encoding: encoding for the HTML file
    """
    try:
        fh = open(file, 'rb')
    except TypeError:
        fh = nullcontext(file)
    except FileNotFoundError:
        fh = None

    if not fh:
        return

    with fh as fh:
        if not encoding:
            encoding = get_html_charset(fh, default='ISO-8859-1')
            fh.seek(0)

        contexts = []
        for event, elem in etree.iterparse(fh, encoding=encoding, html=True, events=('start', 'end')):
            if event == 'start':
                if elem.tag in META_REFRESH_CONTEXT_TAGS:
                    contexts.append(elem.tag)
                    continue

                if (elem.tag == 'meta' and elem.attrib.get('http-equiv', '').lower() == 'refresh'):
                    yield parse_meta_refresh_content(elem.attrib.get('content', ''), contexts)

            elif event == 'end':
                if contexts and elem.tag == contexts[-1]:
                    contexts.pop()
                    continue

                # clean up to save memory
                elem.clear()
                while elem.getprevious() is not None:
                    try:
                        del elem.getparent()[0]
                    except TypeError:
                        # broken html may generate extra root elem
                        break


def get_meta_refresh(file):
    """Retrieve a general redirect-like meta refresh from a file.

    According to spec, the document should take only the first meta refresh.
    This is also more performant. Though many browsers accept multiple meta
    refresh, and usually the last one wins.

    As the document may contain multiple contexted meta refreshes and we cannot
    determine which should apply, we retrieve the first non-contexted one and
    accept it only when its time is 0.

    https://html.spec.whatwg.org/multipage/semantics.html#attr-meta-http-equiv-refresh

    Args:
        file: str, path-like, or file-like bytes object
    """
    for info in iter_meta_refresh(file):
        if info.context:
            continue
        if info.time == 0:
            return info
        break
    return MetaRefreshInfo(time=None, target=None, context=None)


def get_meta_refreshed_file(file):
    """Resolve the meta-refreshed file.

    Returns:
        path of the meta-refreshed file. None if there's no valid meta refresh,
            the meta refresh points to an external resource, or the path of the
            meta refresh points to the input file.

    Raises:
        MetaRefreshCircularError: if the meta refresh is circular
    """
    _file = file
    url_chain = set()

    while True:
        doc_url = urljoin('file:///', pathname2url(file))
        if doc_url in url_chain:
            raise MetaRefreshCircularError('circular meta refresh')

        url_chain.add(doc_url)

        _, target, _ = get_meta_refresh(file)

        if not target:
            return file if file != _file else None

        url = urljoin(doc_url, target)
        urlparts = urlsplit(url)

        # non-file URL
        if urlparts.scheme != 'file':
            return file if file != _file else None

        file = url2pathname(urlsplit(url).path)


#########################################################################
# MAFF manipulation
#########################################################################

MaffPageInfo = namedtuple('MaffPageInfo', ('title', 'originalurl', 'archivetime', 'indexfilename', 'charset'))


def get_maff_pages(zip):
    """Get a list of pages (MaffPageInfo).

    Args:
        zip: path, file-like object, or zipfile.ZipFile
    """
    pages = []
    with nullcontext(zip) if isinstance(zip, zipfile.ZipFile) else zipfile.ZipFile(zip) as zh:
        # get top folders and their content files
        topdirs = {}
        for entry in zh.namelist():
            topdir, sep, p = entry.partition('/')
            topdir = topdirs.setdefault(topdir + sep, [])
            if p:
                topdir.append(entry)

        # get index files
        for topdir in topdirs:
            rdf = topdir + 'index.rdf'
            try:
                with zh.open(rdf, 'r') as fh:
                    meta = parse_maff_index_rdf(fh)
            except Exception:
                pass
            else:
                if meta.indexfilename is not None:
                    pages.append(MaffPageInfo(
                        meta.title,
                        meta.originalurl,
                        meta.archivetime,
                        topdir + meta.indexfilename,
                        meta.charset,
                    ))
                    continue

            for entry in topdirs[topdir]:
                if entry.startswith(topdir + 'index.') and entry != topdir + 'index.rdf':
                    pages.append(MaffPageInfo(
                        None,
                        None,
                        None,
                        entry,
                        None,
                    ))

    return pages


def parse_maff_index_rdf(fh):
    """Read MAFF metadata from the given RDF file handler.
    """
    def load_attr(attr):
        try:
            node = root.find('./RDF:Description/MAF:' + attr, ns)
            return node.attrib['{' + ns['RDF'] + '}' + 'resource']
        except Exception:
            return None

    ns = {
        'MAF': 'http://maf.mozdev.org/metadata/rdf#',
        'NC': 'http://home.netscape.com/NC-rdf#',
        'RDF': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
    }

    text = fh.read().decode('UTF-8')
    root = etree.XML(text)

    return MaffPageInfo(
        load_attr('title'),
        load_attr('originalurl'),
        load_attr('archivetime'),
        load_attr('indexfilename'),
        load_attr('charset'),
    )


#########################################################################
# Encrypt and security
#########################################################################

class Encrypt():
    """Simple hash encryption with salt.
    """
    def md5(self, text, salt=''):
        return hashlib.md5((text + salt).encode('UTF-8')).hexdigest()

    def sha1(self, text, salt=''):
        return hashlib.sha1((text + salt).encode('UTF-8')).hexdigest()

    def sha224(self, text, salt=''):
        return hashlib.sha224((text + salt).encode('UTF-8')).hexdigest()

    def sha256(self, text, salt=''):
        return hashlib.sha256((text + salt).encode('UTF-8')).hexdigest()

    def sha384(self, text, salt=''):
        return hashlib.sha384((text + salt).encode('UTF-8')).hexdigest()

    def sha512(self, text, salt=''):
        return hashlib.sha512((text + salt).encode('UTF-8')).hexdigest()

    def sha3_224(self, text, salt=''):
        return hashlib.sha3_224((text + salt).encode('UTF-8')).hexdigest()

    def sha3_256(self, text, salt=''):
        return hashlib.sha3_256((text + salt).encode('UTF-8')).hexdigest()

    def sha3_384(self, text, salt=''):
        return hashlib.sha3_384((text + salt).encode('UTF-8')).hexdigest()

    def sha3_512(self, text, salt=''):
        return hashlib.sha3_512((text + salt).encode('UTF-8')).hexdigest()

    def plain(self, text, salt=''):
        return text + salt

    def encrypt(self, text, salt='', method='plain'):
        fn = getattr(self, method, None)

        if not callable(fn):
            print(f"Encrypt method {method!r} not implemented, fallback to 'plain'.", file=sys.stderr)
            fn = self.plain

        return fn(text, salt)


encrypt = Encrypt().encrypt
